@@ -25,32 +25,17 @@ MA 02111-1307, USA. */
 #include <netdb.h>
 #include <unistd.h>
 #include <stdio.h>
-#include <errno.h>
 
 #include "libssh/priv.h"
 #include "libssh/ssh2.h"
 #define WINDOWLIMIT 1024
 #define WINDOWBASE 32000
-
-/** defgroup ssh_channel
- * \brief functions that manage a channel
- */
-
-/** \addtogroup ssh_channel
- * @{ */
-
-/** \brief allocate a new channel
- * \param session ssh session
- * \return an allocated channel. As this function doesn't speak with server, it never fails
- */
-CHANNEL *channel_new(SSH_SESSION *session){
+static void channel_default_bufferize(CHANNEL *channel, void *data, int len, int is_stderr);
+static CHANNEL *new_channel(SSH_SESSION *session){
     CHANNEL *channel=malloc(sizeof(CHANNEL));
     memset(channel,0,sizeof(CHANNEL));
     channel->session=session;
-    channel->version=session->version;
-    channel->stdout_buffer=buffer_new();
-    channel->stderr_buffer=buffer_new();
-     if(!session->channels){
+    if(!session->channels){
         session->channels=channel;
         channel->next=channel->prev=channel;
         return channel;
@@ -62,25 +47,25 @@ CHANNEL *channel_new(SSH_SESSION *session){
     return channel;
 }
 
-u32 ssh_channel_new_id(SSH_SESSION *session){
+static u32 channel_new_id(SSH_SESSION *session){
     u32 ret=session->maxchannel;
     session->maxchannel++;
     return ret;
 }
 
-static int channel_open(CHANNEL *channel,char *type_c,int window,
-        int maxpacket,BUFFER *payload){
-    SSH_SESSION *session=channel->session;
+static CHANNEL *channel_open(SSH_SESSION *session,char *type_c,int window,
+int maxpacket,BUFFER *payload){
+    CHANNEL *channel=new_channel(session);
     STRING *type=string_from_char(type_c);
     u32 foo;
     int err;
     packet_clear_out(session);
     buffer_add_u8(session->out_buffer,SSH2_MSG_CHANNEL_OPEN);
-    channel->local_channel=ssh_channel_new_id(session);
+    channel->local_channel=channel_new_id(session);
     channel->local_maxpacket=maxpacket;
     channel->local_window=window;
-    ssh_say(2,"creating a channel %d with %d window and %d max packet\n",
-            channel->local_channel, window,maxpacket);
+    ssh_say(2,"creating a channel %d with %d window and %d max packet\n",channel->local_channel,
+        window,maxpacket);
     buffer_add_ssh_string(session->out_buffer,type);
     buffer_add_u32(session->out_buffer,htonl(channel->local_channel));
     buffer_add_u32(session->out_buffer,htonl(channel->local_window));
@@ -95,9 +80,10 @@ static int channel_open(CHANNEL *channel,char *type_c,int window,
         case SSH2_MSG_CHANNEL_OPEN_CONFIRMATION:
             buffer_get_u32(session->in_buffer,&foo);
             if(channel->local_channel!=ntohl(foo)){
-                ssh_set_error(session,SSH_FATAL,"server answered with sender chan num %d instead of given %d",
+                ssh_set_error(session,SSH_INVALID_DATA,"server answered with sender chan num %d instead of given %d",
                 ntohl(foo),channel->local_channel);
-                return -1;
+                channel_free(channel);
+                return NULL;
                 }
             buffer_get_u32(session->in_buffer,&foo);
             channel->remote_channel=ntohl(foo);
@@ -105,12 +91,12 @@ static int channel_open(CHANNEL *channel,char *type_c,int window,
             channel->remote_window=ntohl(foo);
             buffer_get_u32(session->in_buffer,&foo);
             channel->remote_maxpacket=ntohl(foo);
-            ssh_say(3,"Received a CHANNEL_OPEN_CONFIRMATION for channel %d:%d\n"
-                    ,channel->local_channel,channel->remote_channel);
+            ssh_say(3,"Received a CHANNEL_OPEN_CONFIRMATION for channel %d:%d\n",
+                channel->local_channel,channel->remote_channel);
             ssh_say(3,"Remote window : %ld, maxpacket : %ld\n",
-	        channel->remote_window, channel->remote_maxpacket);
-	        channel->open=1;
-            return 0;
+	    channel->remote_window, channel->remote_maxpacket);
+	    channel->open=1;
+            return channel;
         case SSH2_MSG_CHANNEL_OPEN_FAILURE:
             {
                 u32 code;
@@ -124,16 +110,18 @@ static int channel_open(CHANNEL *channel,char *type_c,int window,
                     channel->local_channel,ntohl(code),error);
                 free(error);
                 free(error_s);
-                return -1;
+                channel_free(channel);
+                return NULL;
             }
         default:
-            ssh_set_error(session,SSH_FATAL,"Received unknown packet %d\n",session->in_packet.type);
-            return -1;
+            ssh_say(0,"Received unknown packet %d\n",session->in_packet.type);
+            channel_free(channel);
+            return NULL;
         }
-    return -1;
+    return NULL;
 }
 
-CHANNEL *ssh_channel_from_local(SSH_SESSION *session,u32 num){
+static CHANNEL *find_local_channel(SSH_SESSION *session,u32 num){
     // we assume we are always the local
     CHANNEL *initchan,*channel;
     initchan=session->channels;
@@ -166,7 +154,7 @@ static CHANNEL *channel_from_msg(SSH_SESSION *session){
         ssh_set_error(session,SSH_FATAL,"Getting channel from message : short read");
         return NULL;
     }
-    channel=ssh_channel_from_local(session,ntohl(chan));
+    channel=find_local_channel(session,ntohl(chan));
     if(!channel)
         ssh_set_error(session,SSH_FATAL,"Server specified invalid channel %d",ntohl(chan));
     return channel;
@@ -214,7 +202,21 @@ static void channel_rcv_data(SSH_SESSION *session,int is_stderr){
     /* what shall we do in this case ? let's accept it anyway */
     if(string_len(str)>channel->local_window)
         ssh_say(0,"Data packet too big for our window(%d vs %d)",string_len(str),channel->local_window);
-    channel_default_bufferize(channel,str->string,string_len(str), is_stderr);
+    if(!is_stderr){
+        /* stdout */
+        if(channel->write_fct){
+            channel->write_fct(channel,str->string,string_len(str),channel->userarg);
+        } else {
+            channel_default_bufferize(channel,str->string,string_len(str),is_stderr);
+        }
+    } else {
+        /* stderr */
+        if(channel->write_err_fct){
+            channel->write_err_fct(channel,str->string,string_len(str),channel->userarg);
+        } else {
+            channel_default_bufferize(channel,str->string,string_len(str),is_stderr);
+        }
+    }
     if(string_len(str)>=channel->local_window)
         channel->local_window-=string_len(str);
     else
@@ -246,16 +248,10 @@ static void channel_rcv_close(SSH_SESSION *session){
     }
     ssh_say(2,"Received close on channel (%d:%d)\n",channel->local_channel,
         channel->remote_channel);
-    if((channel->stdout_buffer && buffer_get_rest_len(channel->stdout_buffer)>0)
-            || (channel->stderr_buffer && buffer_get_rest_len(channel->stderr_buffer)>0))
-        channel->delayed_close=1;
-    else
-        channel->open=0;
+    channel->open=0;
     if(!channel->remote_eof)
         ssh_say(2,"Remote host not polite enough to send an eof before close\n");
     channel->remote_eof=1;
-    /* the remote eof doesn't break things if there was still data into read
-     * buffer because the eof is ignored until the buffer is empty. */
 }
 
 static void channel_rcv_request(SSH_SESSION *session){
@@ -340,7 +336,7 @@ void channel_handle(SSH_SESSION *session, int type){
 /* when data has been received from the ssh server, it can be applied to the known
     user function, with help of the callback, or inserted here */
 /* XXX is the window changed ? */
-void channel_default_bufferize(CHANNEL *channel, void *data, int len, int is_stderr){
+static void channel_default_bufferize(CHANNEL *channel, void *data, int len, int is_stderr){
     ssh_say(3,"placing %d bytes into channel buffer (stderr=%d)\n",len,is_stderr);
     if(!is_stderr){
         /* stdout */
@@ -355,41 +351,24 @@ void channel_default_bufferize(CHANNEL *channel, void *data, int len, int is_std
     }
 }
 
-/** \brief open a session channel (suited for a shell. Not tcp)
- * \param channel an allocated channel (see channel_new())
- * \return SSH_OK on success\n
- * SSH_ERROR on error
- * \see channel_open_forward()
- * \see channel_request_env()
- * \see channel_request_shell()
- * \see channel_request_exec()
- * \warning API changed from 0.11
- */
-int channel_open_session(CHANNEL *channel){
-#ifdef HAVE_SSH1
-    if(channel->session->version==2)
-#endif
-        return channel_open(channel,"session",64000,32000,NULL);
-#ifdef HAVE_SSH1
-    else
-        return channel_open_session1(channel);
-#endif
-}    
 
-/** \brief open a TCP/IP forwarding channel.
- * \param channel an allocated channel (see channel_new())
- * \param remotehost remote host to be connected (host name or IP)
- * \param remoteport remote port
- * \param sourcehost source host (your local computer). It's facultative and for logging purpose
- * \param localport source port (your local computer). It's facultative and for logging purpose
- * \return SSH_ERROR on error\n
- * SSH_OK on success
- * \warning API changed from 0.11
- */
-int channel_open_forward(CHANNEL *channel,char *remotehost, int remoteport, char *sourcehost, int localport){
+/* --8<-- PUBLIC INTERFACE BEGINS HERE -8<-----8< --- */
+
+/* deprecated */
+CHANNEL *open_session_channel(SSH_SESSION *session,int window,int maxpacket){
+    CHANNEL *chan=channel_open(session,"session",window,maxpacket,NULL);
+    return chan;
+}
+
+CHANNEL *channel_open_session(SSH_SESSION *session){
+    return open_session_channel(session,64000,32000);
+}
+
+/* tcpip forwarding */
+CHANNEL *channel_open_forward(SSH_SESSION *session,char *remotehost, int remoteport, char *sourcehost, int localport){
+    CHANNEL *chan;
     BUFFER *payload=buffer_new();
     STRING *str=string_from_char(remotehost);
-    int ret;
     buffer_add_ssh_string(payload,str);
     free(str);
     str=string_from_char(sourcehost);
@@ -397,15 +376,12 @@ int channel_open_forward(CHANNEL *channel,char *remotehost, int remoteport, char
     buffer_add_ssh_string(payload,str);
     free(str);
     buffer_add_u32(payload,htonl(localport));
-    ret=channel_open(channel,"direct-tcpip",64000,32000,payload);
+    chan=channel_open(session,"direct-tcpip",64000,32000,payload);
     buffer_free(payload);
-    return ret;
+    return chan;
 }
 
-/** \brief close and free a channel
- * \param channel channel to free
- * \warning any data unread on channel will be lost
- */
+
 void channel_free(CHANNEL *channel){
     SSH_SESSION *session=channel->session;
     if(session->alive && channel->open)
@@ -429,14 +405,6 @@ void channel_free(CHANNEL *channel){
     free(channel);
 }
 
-/** it doesn't close the channel. You may still read from it but not write. 
- * \brief send an end of file on the channel
- * \param channel channel
- * \return SSH_ERROR on error\n
- * SSH_SUCCESS on success
- * \see channel_close()
- * \see channel_free()
- */
 int channel_send_eof(CHANNEL *channel){
     SSH_SESSION *session=channel->session;
     int ret;
@@ -450,15 +418,6 @@ int channel_send_eof(CHANNEL *channel){
     return ret;
 }
 
-/** It sends an end of file and then closes the channel. You won't be able
- * to recover any data the server was going to send or was in buffers.
- * \brief close a channel
- * \param channel channel
- * \return SSH_ERROR on error\n
- * SSH_SUCCESS on success
- * \see channel_free()
- * \see channel_eof()
- */
 int channel_close(CHANNEL *channel){
     SSH_SESSION *session=channel->session;
     int ret=0;
@@ -477,14 +436,8 @@ int channel_close(CHANNEL *channel){
     return ret;
 }
 
-/** \brief blocking write on channel
- * \param channel channel
- * \param data pointer to data to write
- * \param len length of data
- * \return number of bytes written on success\n
- * SSH_ERROR on error
- * \see channel_read()
- */
+/* Blocking write */
+/* The exact len is written */
 int channel_write(CHANNEL *channel ,void *data,int len){
     SSH_SESSION *session=channel->session;
     int effectivelen;
@@ -494,14 +447,6 @@ int channel_write(CHANNEL *channel ,void *data,int len){
             " after EOF was sent",channel->local_channel,channel->remote_channel);
         return -1;
     }
-    if(!channel->open || channel->delayed_close){
-        ssh_set_error(session,SSH_REQUEST_DENIED,"Remote channel is closed");
-        return -1;
-    }
-#ifdef HAVE_SSH1
-    if(channel->version==1)
-        return channel_write1(channel,data,len);
-#endif
     while(len >0){
         if(channel->remote_window<len){
 	    ssh_say(2,"Remote window is %d bytes. going to write %d bytes\n",
@@ -524,51 +469,15 @@ int channel_write(CHANNEL *channel ,void *data,int len){
         ssh_say(2,"channel_write wrote %d bytes\n",effectivelen);
         channel->remote_window-=effectivelen;
         len -= effectivelen;
-        data = (char*)data + effectivelen;
+        data+=effectivelen;
     }
     return origlen;
 }
 
-/** \brief returns if the channel is open or not
- * \param channel channel
- * \return 0 if channel is closed, nonzero otherwise
- * \see channel_is_closed()
- */
 int channel_is_open(CHANNEL *channel){
-    return (channel->open!=0 && channel->session->alive);
+    return (channel->open!=0);
 }
 
-/** \brief returns if the channel is closed or not
- * \param channel channel
- * \return 0 if channel is opened, nonzero otherwise
- * \see channel_is_open()
- */
-
-int channel_is_closed(CHANNEL *channel){
-    return (channel->open==0 || !channel->session->alive);
-}
-
-/** \brief returns if the remote has sent an EOF
- * \param channel channel
- * \return 0 if there is no EOF, nonzero otherwise
- */
-int channel_is_eof(CHANNEL *channel){
-    if((channel->stdout_buffer && buffer_get_rest_len(channel->stdout_buffer)
-                >0) || (channel->stderr_buffer && buffer_get_rest_len(
-                    channel->stderr_buffer)>0))
-        return 0;
-    return (channel->remote_eof!=0);
-}
-
-/** \brief put the channel into nonblocking mode
- * \param channel channel
- * \param blocking boolean for blocking or nonblocking
- * \bug This functionnality is still under development and
- * doesn't work correctly
- */
-void channel_set_blocking(CHANNEL *channel, int blocking){
-    channel->blocking=blocking;
-}
 
 static int channel_request(CHANNEL *channel,char *request, BUFFER *buffer,int reply){
     STRING *request_s=string_from_char(request);
@@ -599,24 +508,10 @@ static int channel_request(CHANNEL *channel,char *request, BUFFER *buffer,int re
     return err;
 }
 
-/** \brief requests a pty with a specific type and size
- * \param channel channel
- * \param terminal terminal type ("vt100, xterm,...")
- * \param col number of cols
- * \param row number of rows
- * \return SSH_SUCCESS on success\n
- * SSH_ERROR on error
- */
 int channel_request_pty_size(CHANNEL *channel, char *terminal, int col, int row){
-    STRING *term;
-    BUFFER *buffer;
+    STRING *term=string_from_char(terminal);
+    BUFFER *buffer=buffer_new();
     int err;
-#ifdef HAVE_SSH1
-    if(channel->version==1)
-        return channel_request_pty_size1(channel,terminal, col, row);
-#endif
-    term=string_from_char(terminal);
-    buffer=buffer_new();
     buffer_add_ssh_string(buffer,term);
     buffer_add_u32(buffer,htonl(col));
     buffer_add_u32(buffer,htonl(row));
@@ -631,32 +526,13 @@ int channel_request_pty_size(CHANNEL *channel, char *terminal, int col, int row)
     return err;
 }
 
-/** \brief requests a pty
- * \param channel channel
- * \see channel_request_pty_size()
- * \return SSH_SUCCESS on success\n
- * SSH_ERROR on error
- */
 int channel_request_pty(CHANNEL *channel){
     return channel_request_pty_size(channel,"xterm",80,24);
 }
 
-/** \brief change the size of the terminal associated to a channel
- * \param channel channel
- * \param cols new number of cols
- * \param rows new number of rows
- * \warning Do not call it from a signal handler if you are not
- * sure any other libssh function using the same channel/session
- * is running at same time (not 100% threadsafe)
- */
 int channel_change_pty_size(CHANNEL *channel,int cols,int rows){
-    BUFFER *buffer;
+    BUFFER *buffer=buffer_new();
     int err;
-#ifdef HAVE_SSH1
-    if(channel->version==1)
-        return channel_change_pty_size1(channel,cols,rows);
-#endif
-    buffer=buffer_new();
     //buffer_add_u8(buffer,0);
     buffer_add_u32(buffer,htonl(cols));
     buffer_add_u32(buffer,htonl(rows));
@@ -666,28 +542,12 @@ int channel_change_pty_size(CHANNEL *channel,int cols,int rows){
     buffer_free(buffer);
     return err;
 }    
-
-/** \brief requests a shell
- * \param channel
- * \returns SSH_SUCCESS on success\n
- * SSH_ERROR on error
- */
+    
 int channel_request_shell(CHANNEL *channel){
-#ifdef HAVE_SSH1
-    if(channel->version==1)
-        return channel_request_shell1(channel);
-#endif
-    return channel_request(channel,"shell",NULL,1);
+    int err=channel_request(channel,"shell",NULL,1);
+    return err;
 }
 
-/** \brief requests a subsystem (for example sftp)
- * \param channel channel
- * \param system subsystem to request (for example sftp)
- * \return SSH_SUCCESS on success\n
- * SSH_ERROR on error
- * \warning you normally don't have to call it to have sftp
- * \see sftp_new()
- */
 int channel_request_subsystem(CHANNEL *channel, char *system){
     BUFFER* buffer=buffer_new();
     int ret;
@@ -703,14 +563,7 @@ int channel_request_sftp( CHANNEL *channel){
     return channel_request_subsystem(channel, "sftp");
 }
 
-/** \brief set the environement variables
- * \param channel channel
- * \param name name of the variable
- * \param value value
- * \return SSH_SUCCESS on success\n
- * SSH_ERROR on error
- * \warning some environement variables may be refused by security
- * */
+
 int channel_request_env(CHANNEL *channel,char *name, char *value){
     BUFFER *buffer=buffer_new();
     int ret;
@@ -725,22 +578,9 @@ int channel_request_env(CHANNEL *channel,char *name, char *value){
     return ret;
 }
 
-/** it's similar to sh -c "command"
- * \brief run a shell command without an interactive shell
- * \param channel channel
- * \param cmd command to execute (by ex. "ls ~/ -al | grep -i reports")
- * \return SSH_SUCCESS on success\n
- * SSH_ERROR on error
- * \see channel_request_shell()
- */
 int channel_request_exec(CHANNEL *channel, char *cmd){
-    BUFFER *buffer;
+    BUFFER *buffer=buffer_new();
     int ret;
-#ifdef HAVE_SSH1
-    if(channel->version==1)
-        return channel_request_exec1(channel, cmd);
-#endif
-    buffer=buffer_new();
     STRING *command=string_from_char(cmd);
     buffer_add_ssh_string(buffer,command);
     free(command);
@@ -749,32 +589,42 @@ int channel_request_exec(CHANNEL *channel, char *cmd){
     return ret;
 }
 
-/* TODO : fix the delayed close thing */
-/* TODO : fix the blocking behaviours */
+int channel_set_write_handler(CHANNEL *chan,
+        void (*write_fct)(CHANNEL *channel, void *data, int len, void *userdefined),void *user){
+    chan->write_fct=write_fct;
+    chan->userarg=user;
+    return 0;
+}
+
+int channel_set_stderr_write_handler(CHANNEL *chan,
+        void (*write_err_fct)(CHANNEL *channel, void *data, int len, void *userdefined),void *user){
+    chan->write_err_fct=write_err_fct;
+    chan->userarg=user;
+    return 0;
+}
+
+
 /* reads into a channel and put result into buffer */
 /* returns number of bytes read, 0 if eof or such and -1 in case of error */
 /* if bytes != 0, the exact number of bytes are going to be read */
-/** \brief reads data from a channel
- * \param channel channel
- * \param buffer buffer which will get the data
- * \param bytes number of bytes to be read. If it is bigger
- * than 0, the exact size will be read, else (bytes=0) it will return
- * once anything is available
- * \param is_stderr boolean value to mark reading from the stderr flow.
- * \return number of bytes read\n
- * 0 on end of file\n
- * SSH_ERROR on error
- */
 int channel_read(CHANNEL *channel, BUFFER *buffer,int bytes,int is_stderr){
     BUFFER *stdbuf=NULL;
     int len;
     buffer_reinit(buffer);
     /* maybe i should always set a buffer to avoid races between channel_default_bufferize and channel_read */
-    if(is_stderr)
+    if(channel->write_fct){
+        ssh_set_error(channel->session,SSH_INVALID_REQUEST,"Specified channel hasn't got a default buffering system\n");
+        return -1;
+    }
+    if(is_stderr){
+        if(!channel->stderr_buffer)
+            channel->stderr_buffer=buffer_new();
         stdbuf=channel->stderr_buffer;
-    else 
+    } else {
+        if(!channel->stdout_buffer)
+            channel->stdout_buffer=buffer_new();
         stdbuf=channel->stdout_buffer;
-    
+    }
     /* block reading if asked bytes=0 */
     while((buffer_get_rest_len(stdbuf)==0) || (buffer_get_rest_len(stdbuf) < bytes)){
         if(channel->remote_eof && buffer_get_rest_len(stdbuf)==0)
@@ -799,50 +649,31 @@ int channel_read(CHANNEL *channel, BUFFER *buffer,int bytes,int is_stderr){
     return buffer_get_len(buffer);
 }
 
-/** \brief polls the channel for data to read
- * \param channel channel
- * \param is_stderr boolean to select the stderr stream
- * \return number of bytes available for reading\n
- * 0 if nothing is available\n
- * SSH_ERROR on error
- * \warning don't forget to check for EOF as it would 
- * return 0 here
- * \see channel_is_eof()
- */
+/* returns the number of bytes available, 0 if nothing is currently available, -1 if error */
 int channel_poll(CHANNEL *channel, int is_stderr){
     BUFFER *buffer;
-    int r=0;
-    if(is_stderr)
+    if(is_stderr){
         buffer=channel->stderr_buffer;
-    else 
+        if(!buffer)
+            buffer=channel->stderr_buffer=buffer_new();
+    } else {
         buffer=channel->stdout_buffer;
-    
-    while(buffer_get_rest_len(buffer)==0 && !channel->remote_eof){
-        r=ssh_handle_packets(channel->session);
-        if(r<=0)
-            return r;
+        if(!buffer)
+            buffer=channel->stdout_buffer=buffer_new();
+    }
+    while(buffer_get_len(buffer)==0){
+        if(ssh_fd_poll(channel->session)){
+            if(packet_read(channel->session)||packet_translate(channel->session))
+                return -1;
+            packet_parse(channel->session);
+        } else
+            return 0; /* nothing is available has said fd_poll */
     }
-    if(channel->remote_eof)
-        return 1;
     return buffer_get_len(buffer);
 }
 
 /* nonblocking read on the specified channel. it will return <=len bytes of data read
  atomicly. */
-/** This read will make a nonblocking read (unlike channel_read()) and won't force you
- * to deal with BUFFER's
- * \brief nonblocking read
- * \param channel channel
- * \param dest pointer to destination for data
- * \param len maximum length of data to be read
- * \param is_stderr boolean to select the stderr stream
- * \return number of bytes read\n
- * 0 if nothing is available\n
- * SSH_ERROR on error
- * \warning don't forget to check for EOF as it would 
- * return 0 here
- * \see channel_is_eof()
- */
 int channel_read_nonblocking(CHANNEL *channel, char *dest, int len, int is_stderr){
     int to_read=channel_poll(channel,is_stderr);
     int lu;
@@ -858,176 +689,3 @@ int channel_read_nonblocking(CHANNEL *channel, char *dest, int len, int is_stder
     buffer_free(buffer);
     return lu;
 }
-
-/** \brief recover the session in which belong a channel
- * \param channel channel
- * \return the session pointer
- */
-SSH_SESSION *channel_get_session(CHANNEL *channel){
-    return channel->session;
-}
-
-/* This function acts as a meta select. */
-/* first, channels are analyzed to seek potential can-write or can-read ones. */
-/* Then, if no channel has been elected, it goes in a loop with the posix select(2) */
-/* this is made in two parts : Protocol select and Network select */
-
-/* the protocol select does not use the network functions at all */
-
-static int channel_protocol_select(CHANNEL **rchans, CHANNEL **wchans, CHANNEL **echans,
-                            CHANNEL **rout, CHANNEL **wout, CHANNEL **eout){
-    CHANNEL *chan;
-    int i,j;
-    j=0;
-    for(i=0;rchans[i];++i){
-        chan=rchans[i];
-        while(chan->open && chan->session->data_to_read){
-            ssh_handle_packets(chan->session);
-        }
-        if( (chan->stdout_buffer && buffer_get_len(chan->stdout_buffer)>0) ||
-             (chan->stderr_buffer && buffer_get_len(chan->stderr_buffer)>0) ||
-            chan->remote_eof){
-            rout[j]=chan;
-            ++j;
-        }
-    }
-    rout[j]=NULL;
-    j=0;
-    for(i=0;wchans[i];++i){
-        chan=wchans[i];
-        /* it's not our business to seek if the file descriptor is writable */
-        if(chan->session->data_to_write && chan->open && (chan->remote_window>0)){
-            wout[j]=chan;
-            ++j;
-        }
-    }
-    wout[j]=NULL;
-    j=0;
-    for(i=0;echans[i];++i){
-        chan=echans[i];
-        if(chan->session->fd==-1 || !chan->open || chan->session->data_except){
-            eout[j]=chan;
-            ++j;
-        }
-    }
-    eout[j]=NULL;
-    return 0;
-}
-
-/* just count number of pointers in the array */
-static int count_ptrs(CHANNEL **ptrs){
-    int c;
-    for(c=0;ptrs[c];++c)
-        ;
-    return c;
-}
-
-/** the list of pointers are then actualized and will only contain pointers to
- * channels that are respectively readable, writable or have an exception to trap
- * \brief act as the standard select(2) for channels
- * \param readchans a NULL pointer or an array of channel pointers, finished by a NULL
- * \param writechans a NULL pointer or an array of channel pointers, finished by a NULL
- * \param exceptchans a NULL pointer or an array of channel pointers, finished by a NULL
- * \param timeout timeout as defined by select(2)
- * \return SSH_SUCCESS operation successful\n
- * SSH_EINTR select(2) syscall was interrupted, relaunch the function
- */
-int channel_select(CHANNEL **readchans, CHANNEL **writechans, CHANNEL **exceptchans, struct 
-        timeval * timeout){
-    fd_set rset;
-    fd_set wset;
-    fd_set eset;
-    CHANNEL *dummy=NULL;
-    CHANNEL **rchans, **wchans, **echans;
-    int fdmax=-1;
-    int i,fd;
-    int r;
-    /* don't allow NULL pointers */
-    if(!readchans)
-        readchans=&dummy;
-    if(!writechans)
-        writechans=&dummy;
-    if(!exceptchans)
-        exceptchans=&dummy;
-    if(!readchans[0] && !writechans[0] && !exceptchans[0]){
-        /* no channel to poll ?? go away ! */
-        return 0;
-    }
-    /* prepare the outgoing temporary arrays */
-    rchans=malloc(sizeof(CHANNEL *) * (count_ptrs(readchans)+1));
-    wchans=malloc(sizeof(CHANNEL *) * (count_ptrs(writechans)+1));
-    echans=malloc(sizeof(CHANNEL *) * (count_ptrs(exceptchans)+1));
-
-    /* first, try without doing network stuff */
-    /* then, select and redo the networkless stuff */
-    do {
-        channel_protocol_select(readchans,writechans,exceptchans,rchans,wchans,echans);
-        if(rchans[0]||wchans[0]||echans[0]){
-            /* we've got one without doing any select */
-            /* overwrite the begining arrays */
-            memcpy(readchans,rchans, (count_ptrs(rchans)+1)*sizeof(CHANNEL *));
-            memcpy(writechans,wchans, (count_ptrs(wchans)+1)*sizeof(CHANNEL *));
-            memcpy(exceptchans,echans, (count_ptrs(echans)+1)*sizeof(CHANNEL *));
-            free(rchans);
-            free(wchans);
-            free(echans);
-            return 0;
-        }
-        ssh_say(3,"doing a select for the different channels\n");
-        /* since we verified the invalid fd cases into the networkless select,
-        we can be sure all fd are valid ones */
-        FD_ZERO(&rset);
-        FD_ZERO(&wset);
-        FD_ZERO(&eset);
-        for(i=0;readchans[i];++i){
-            fd=readchans[i]->session->fd;
-            if(!FD_ISSET(fd,&rset)){
-                FD_SET(fd,&rset);
-                if(fd>=fdmax)
-                    fdmax=fd+1;
-            }
-        }
-        for(i=0;writechans[i];++i){
-            fd=writechans[i]->session->fd;
-            if(!FD_ISSET(fd,&wset)){
-                FD_SET(fd,&wset);
-                if(fd>=fdmax)
-                    fdmax=fd+1;
-            }
-        }
-        for(i=0;exceptchans[i];++i){
-            fd=exceptchans[i]->session->fd;
-            if(!FD_ISSET(fd,&eset)){
-                FD_SET(fd,&eset);
-                if(fd>=fdmax)
-                    fdmax=fd+1;
-            }
-        }
-        /* here we go */
-        r=select(fdmax,&rset,&wset,&eset,timeout);
-        /* leave if select was interrupted */
-        if(r==EINTR){
-            free(rchans);
-            free(wchans);
-            free(echans);
-            return SSH_EINTR;
-        }
-        for(i=0;readchans[i];++i){
-            if(FD_ISSET(readchans[i]->session->fd,&rset))
-                readchans[i]->session->data_to_read=1;
-        }
-        for(i=0;writechans[i];++i){
-            if(FD_ISSET(writechans[i]->session->fd,&wset))
-                writechans[i]->session->data_to_write=1;
-        }
-        for(i=0;exceptchans[i];++i){
-            if(FD_ISSET(exceptchans[i]->session->fd,&eset))
-                exceptchans[i]->session->data_except=1;
-        }
-    } while(1); /* return to do loop */
-    /* not reached */
-    return 0;
-}
-
-/** @} */
-
